// Outputs a gpx file containing a track that represents the geodetic path from point1 to point2 
// Released as Public Domain.
// By Michele Zanin http://www.miczanin.it
// Requires GeographicLib http://geographiclib.sourceforge.net 
// Please compile with 'g++ -o geodesic geodesic.cpp -lGeographic'

#include <iostream>
#include <GeographicLib/Geodesic.hpp>
#include <GeographicLib/GeodesicLine.hpp>

int main(int argc, char** argv) 
{
  try
    {
      if (argc < 5) { std::cerr << argv[0] << " lat1 lon1 lat2 lon2" << std::endl; return -2;}
      double lat1 = atof(argv[1]); double lon1 = atof(argv[2]); // point 1 (no check on latlon range)
      double lat2 = atof(argv[3]); double lon2 = atof(argv[4]); // point 2 (no check on latlon range)
      double ds0 = 500.0e3; // Nominal distance between points (default = 500km)
      if (argc > 5) ds0 = atof(argv[5]);
      if (ds0 < 1000.0) ds0 = 1000.0; // min ds0 is 1km

      // initialization
      const GeographicLib::Geodesic& geod = GeographicLib::Geodesic::WGS84;
      double s12, azi1, azi2, a12;
      a12 = geod.Inverse(lat1, lon1, lat2, lon2, s12, azi1, azi2);
      const GeographicLib::GeodesicLine line = geod.Line(lat1, lon1, azi1);
      int num = int(ceil(s12 / ds0)); // The number of intervals
      double ds = s12 / num; // actual distance between points

      // gpx header creation
      std::string ind = "  ";
      int part = 1;
      std::cout << "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" << std::endl << std::endl;
      std::cout << "<gpx version=\"1.1\" creator=\"geodetic - http://www.miczanin.it\">" << std::endl;
      std::cout << ind << "<trk>" << std::endl << ind << ind << "<name>\"Geodetic line from (";
      std::cout << lat1 << "," << lon1 << ") to (" << lat2 << "," << lon2 << "), Part " << part << "\"</name>" << std::endl;
      std::cout << ind << ind << "<trkseg>" << std::endl;

      // writing all points
      double lat, lon, latx, lonx, latp = lat1, lonp = lon1;
      std::cout << ind << ind << ind << "<trkpt lat=\"" << lat1 << "\" lon=\"" << lon1 << "\"></trkpt>" << std::endl;
      for (int i = 1; i <= num; ++i) 
	{
	  line.Position(i * ds, lat, lon); // compute current point
	  
	  // detect wrap-around
	  lonx = -1000.0;
	  if ((lonp > 140.0) && (lon < -140.0)) { lonx = 180.0; latx = latp + (lat-latp) * (lonx-lonp) / (lon+360.0-lonp);}
	  else if ((lonp < -140.0) && (lon > 140.0)) { lonx = -180.0; latx = latp + (lat-latp) * (lonx-lonp) / (lon-360.0-lonp);}
	  if (lonx > -1000.0) // there is a wrap-around (compute extreme point, close track, open another one)
	    {
	      std::cout << ind << ind << ind << "<trkpt lat=\"" << latx << "\" lon=\"" << lonx << "\"></trkpt>" << std::endl;
	      std::cout << ind << ind << "</trkseg>" << std::endl << ind << "</trk>" << std::endl;
	      std::cout << ind << "<trk>" << std::endl << ind << ind << "<name>\"Geodetic line from (";
	      std::cout << lat1 << "," << lon1 << ") to (" << lat2 << "," << lon2 << "), Part " << ++part << "\"</name>" << std::endl;
	      std::cout << ind << ind << "<trkseg>" << std::endl;
	      std::cout << ind << ind << ind << "<trkpt lat=\"" << latx << "\" lon=\"" << (lonx == -180.0 ? 180.0 : -180.0) << "\"></trkpt>" << std::endl;
	    }

	  // write current point
	  std::cout << ind << ind << ind << "<trkpt lat=\"" << lat << "\" lon=\"" << lon << "\"></trkpt>" << std::endl;

	  latp = lat; lonp = lon;
	}

      // close gpx file
      std::cout << ind << ind << "</trkseg>" << std::endl;
      std::cout << ind << "</trk>" << std::endl << "</gpx>" << std::endl;
    }
  catch (const std::exception& e) 
    {
      std::cerr << "Caught exception: " << e.what() << "\n";
      return 1;
    }

  return 0;
}
